/*----------------------------------------------------------------------
  File    : rbft.c
  Contents: radial basis function network training
  Author  : Christian Borgelt
  History : 2004.08.14 file created from file mlpt.c
            2004.09.04 reading of cluster set as initialization added
            2007.01.10 usage of attribute maps changed
            2007.02.14 adapted to modified module tabread
            2007.03.16 table and matrix version combined (option -M)
            2007.07.09 problem with membership normalization fixed
            2007.10.10 evaluation of attribute directions added
            2008.04.08 adapted to modified cluster module
            2013.08.16 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2018.10.17 extended functions (attset, attmap) removed
            2019.11.15 adapted to changed interface of mat_shuffle()
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_MATRIX
#define TAB_MATRIX
#endif
#include "table.h"
#ifndef RBF_PARSE
#define RBF_PARSE
#endif
#include "rbf.h"
#include "random.h"
#include "params.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "rbft"
#define DESCRIPTION "radial basis function network training"
#define VERSION     "version 2.4 (2021.09.01)         " \
                    "(c) 2004-2019   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_TPLCNT    (-10)       /* data file is empty */
#define E_ATTCNT    (-11)       /* no usable attributes found */
#define E_INPCNT    (-12)       /* no usable inputs found */
#define E_UNKTRG    (-13)       /* missing target attribute */
#define E_MULTRG    (-14)       /* multiple target attributes */
#define E_MODE      (-15)       /* invalid init. mode or update meth. */
#define E_INIPAR    (-16)       /* invalid radial function parameter */
#define E_UNITS     (-17)       /* invalid number of units */
#define E_METHOD    (-18)       /* invalid weight update method */
#define E_MODIFY    (-19)       /* invalid weight update modifier */
#define E_LRATE     (-20)       /* invalid learning rate */
#define E_LPARAM    (-21)       /* invalid learning parameter */
#define E_MOMENT    (-22)       /* invalid momentum coefficient */
#define E_EPOCHS    (-23)       /* invalid number of epochs */

#define INPUT      "input"
#define HIDDEN     "hidden"
#define OUTPUT     "output"

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- mode information --- */
  int  code;                    /* code of mode */
  char *name;                   /* name of mode */
  char *desc;                   /* description */
} MODEINFO;                     /* (mode information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "data file is empty",
  /* E_ATTCNT  -11 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -12 */  "no (usable) inputs (need at least 1)",
  /* E_UNKTRG  -13 */  "unknown target attribute '%s'",
  /* E_MULTRG  -14 */  "multiple target attributes found",
  /* E_MODE    -15 */  "unknown mode %s",
  /* E_INIPAR  -16 */  "invalid initialization parameter %g",
  /* E_UNITS   -17 */  "invalid number of %s units",
  /* E_METHOD  -18 */  "invalid parameter update method: %s",
  /* E_MODIFY  -19 */  "invalid parameter update method: %s",
  /* E_LRATE   -20 */  "invalid learning rate: %g",
  /* E_LPARAM  -21 */  "invalid learning parameter: %g",
  /* E_MOMENT  -22 */  "invalid momentum coefficient: %g",
  /* E_EPOCHS  -23 */  "invalid number of epochs: %"DIMID_FMT,
  /*           -24 */  "unknown error",
};

static const MODEINFO initab[] = { /* table of initialization modes */
  { CLS_CENTER,    "center",
                   "center of the data space"                         },
  { CLS_CENTER,    "ctr",                                        NULL },
  { CLS_CENTER,    "c",                                          NULL },
  { CLS_UNIFORM,   "uniform",
                   "samples from a uniform distribution"              },
  { CLS_UNIFORM,   "unif",                                       NULL },
  { CLS_UNIFORM,   "uni",                                        NULL },
  { CLS_UNIFORM,   "u",                                          NULL },
  { CLS_DIAG,      "diag",
                   "points on the diagonal of the data space"         },
  { CLS_DIAG,      "d",                                          NULL },
  { CLS_LATIN,     "latin",
                   "latin hypercube sampling"                         },
  { CLS_LATIN,     "lhcs",                                       NULL },
  { CLS_LATIN,     "lhc",                                        NULL },
  { CLS_LATIN,     "lhs",                                        NULL },
  { CLS_LATIN,     "l",                                          NULL },
  { CLS_POINTS,    "points",
                   "data points chosen uniformly at random"           },
  { CLS_POINTS,    "pts",                                        NULL },
  { CLS_POINTS,    "p",                                          NULL },
  { CLS_MAXIMIN,   "maximin",
                   "choice by maximum minimum distance"               },
  { CLS_MAXIMIN,   "maxmin",                                     NULL },
  { CLS_MAXIMIN,   "max",                                        NULL },
  { CLS_MAXIMIN,   "m",                                          NULL },
  { CLS_KMPP,      "kmeans++",
                   "sampling from d^2 distribution"                   },
  { CLS_KMPP,      "km++",                                       NULL },
  { CLS_KMPP,      "kmpp",                                       NULL },
  { CLS_KMPP,      "kmeanspp",                                   NULL },
  { CLS_KMPP,      "cmeans++",                                   NULL },
  { CLS_KMPP,      "kc++",                                       NULL },
  { CLS_KMPP,      "cmpp",                                       NULL },
  { CLS_KMPP,      "cmeanspp",                                   NULL },
  { CLS_KMPP,      "k",                                          NULL },
  { CLS_KMPP,      "+",                                          NULL },
  { CLS_KMPP,      "++",                                         NULL },
  { CLS_QKMPP,     "qkmeans++",
                   "sampling from quantile range of d^2 distribution" },
  { CLS_QKMPP,     "qkm++",                                      NULL },
  { CLS_QKMPP,     "qkmpp",                                      NULL },
  { CLS_QKMPP,     "qkmeanspp",                                  NULL },
  { CLS_QKMPP,     "q",                                          NULL },
  { CLS_AFKMC2,    "afkmc2",
                   "Markov chain based sampling from d^2"             },
  { CLS_AFKMC2,    "mc2",                                        NULL },
  { CLS_AFKMC2,    "a",                                          NULL },
  { CLS_SAFKMC2,   "safkmc2",
                   "simplified Markov chain based sampling"           },
  { CLS_SAFKMC2,   "smc2",                                       NULL },
  { CLS_SAFKMC2,   "s",                                          NULL },
  { -1,            NULL,        /* sentinel */                   NULL },
};

static const MODEINFO updtab[] = { /* table of update methods */
  { RBF_STANDARD,  "bkprop",    "standard backpropagation"            },
  { RBF_STANDARD,  "bkp",                                        NULL },
  { RBF_STANDARD,  "b",                                          NULL },
  { RBF_ADAPTIVE,  "supersab",  "super self-adaptive backpropagation" },
  { RBF_ADAPTIVE,  "sab",                                        NULL },
  { RBF_ADAPTIVE,  "sa",                                         NULL },
  { RBF_ADAPTIVE,  "s",                                          NULL },
  { RBF_RESILIENT, "rprop",     "resilient backpropagation"           },
  { RBF_RESILIENT, "rp",                                         NULL },
  { RBF_RESILIENT, "r",                                          NULL },
  { RBF_QUICK,     "quick",     "quick backpropagation"               },
  { RBF_QUICK,     "q",                                          NULL },
  { RBF_MANHATTAN, "manhattan", "manhattan training"                  },
  { RBF_MANHATTAN, "mh",                                         NULL },
  { RBF_MANHATTAN, "m",                                          NULL },
  { -1,            NULL,        NULL   /* sentinel */                 },
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain descriptions) */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static TABREAD *tread  = NULL;  /* table reader */
static TABLE   *table  = NULL;  /* table  of training data points */
static MATRIX  *data   = NULL;  /* matrix of training data points */
static MATRIX  *ctrs   = NULL;  /* cluster centers for hard c-means */
static CLSET   *clset  = NULL;  /* cluster set */
static RBFNET  *rbf    = NULL;  /* radial basis function network */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (rbf)    rbf_delete(rbf, 0);   \
  if (ctrs)   mat_delete(ctrs);     \
  if (clset)  cls_delete(clset);    \
  if (data)   mat_delete(data);     \
  if (table)  tab_delete(table, 0); \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on methods */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate startup message */
  printf("list of initialization modes (option -i#)\n");
  printf("(choice of initial positions of the cluster centers;\n");
  printf("a random offset can be added with the option -w#):\n");
  printf("  name        initialization mode\n");
  for (i = 0; initab[i].name; i++)
    if (initab[i].desc)         /* list of initialization modes */
      printf("  %-10s  %s\n", initab[i].name, initab[i].desc);
  printf("list of parameter update methods (option -a#)\n");
  printf("  name        parameter update method\n");
  for (i = 0; updtab[i].name; i++)
    if (updtab[i].desc)         /* list of update method */
      printf("  %-10s  %s\n", updtab[i].name, updtab[i].desc);
  printf("For method bkprop a momentum term and for methods supersab "
         "and rprop\na growth and a shrink factor and a minimum and "
         "a maximum value for\nthe change/learning rate may be "
         "specified (options -g# and -z#).\n");
  exit(0);                      /* describe the weight update methods */
}  /* help() */                 /* and abort the program */

/*--------------------------------------------------------------------*/

static double getsse (RBFNET *rbf, MATRIX *data)
{                               /* --- determine network error */
  DIMID  inpcnt;                /* number of input units */
  DIMID  i;                     /* loop variable for patterns */
  double *vec;                  /* to traverse the training patterns */
  double sse;                   /* sum of (squared) errors */

  inpcnt = rbf_inpcnt(rbf);     /* get the number of input units */
  for (sse = 0, i = mat_rowcnt(data); --i >= 0; ) {
    vec = mat_row(data, i);     /* traverse the training patterns */
    rbf_exec(rbf, vec, NULL);   /* and execute the neural network*/
    sse += rbf_error(rbf, vec +inpcnt);
  }                             /* sum the squared errors */
  return sse;                   /* return the computed sum of errors */
}  /* getsse() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s, *end;             /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  network file */
  CCHAR   *fn_dom  = NULL;      /* name of domain file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_rbf  = NULL;      /* name of output network file */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *trgname = NULL;      /* target attribute name */
  CCHAR   *ininame = "points";  /* name of initialization mode */
  CCHAR   *updname = "bkprop";  /* name of update method */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL;/* table read mode */
  DIMID   inpcnt   = 0;         /* number of input  units */
  DIMID   outcnt   = 1;         /* number of output units */
  DIMID   hidcnt   = 2;         /* number of hidden units */
  int     type     = CLS_CENTER;/* cluster type */
  double  rfnps[]  = {2, 1};    /* radial function parameters */
  int     gauss    = 1;         /* flag for Gaussian function */
  int     owrite   = 0;         /* overwrite parameters with options */
  int     inimode  = 0;         /* initialization mode */
  double  range    = 0;         /* range for random offsets */
  double  radius   = 1.0;       /* initial cluster size (radius) */
  int     norm     = 1;         /* whether to normalize inputs */
  double  size     = 100;       /* init.: section size/chain length */
  double  trim     = 0;         /* init.: largest distance to trim */
  int     method   = 0;         /* learning method */
  int     clsupd   = CLS_NONE;  /* update modifiers */
  DIMID   epochs   = 1000, e;   /* number of epochs to train */
  DIMID   update   = 0,    u;   /* number of patterns between updates */
  DIMID   verbose  = 0,    v;   /* flag for verbose output */
  int     shuffle  = 1;         /* whether to shuffle data points */
  double  lrates[] = {0.02,0.01,0.001};  /* learning rates */
  double  term     = 0.0;       /* maximum sse for termination */
  double  delta    = 0.0;       /* maximum sse change for termination */
  int     recomp   = 1;         /* recompute error at end of epoch */
  double  moment   = 0.0;       /* momentum coefficient */
  double  growth   = 1.2;       /* growth factor for learning rate */
  double  shrink   = 0.7;       /* shrink factor for learning rate */
  double  minchg   = 0.0;       /* minimal change/learning rate */
  double  maxchg   = 2.0;       /* maximal change/learning rate */
  double  decay    = 0.0;       /* weight decay */
  double  regps[]  = {0,1,1,0,0};  /* regularization parameters */
  int     maxlen   = 0;         /* maximal output line length */
  long    seed;                 /* seed for random numbers */
  double  sse, sseold;          /* (old) sum of (squared) errors */
  double  *vec;                 /* to traverse the training patterns */
  ATTID   a, c;                 /* number of attributes */
  DIMID   m;                    /* number of inputs/dimensions */
  DIMID   n, r;                 /* number of data points/vectors */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] domfile [-M|[-d|-h hdrfile]] "
                    "tabfile rbffile [infile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-!       print a list of init. modes and update methods\n");
    printf("-M       input is a numerical matrix            "
                    "(default: input is a table)\n");
    printf("-o#      target attribute/number of outputs     "
                    "(default: last attribute/1)\n");
    printf("-c#      number of hidden units/clusters        "
                    "(default: %"DIMID_FMT")\n", hidcnt);
    printf("-G       use Cauchy radial function             "
                    "(default: Gaussian function)\n");
    printf("-p#:#    radial function parameters             "
                    "(default: %g:%g)\n", rfnps[0], rfnps[1]);
    printf("-v       adaptable variances                    "
                    "(default: unit variances)\n");
    printf("-V       adaptable covariances                  "
                    "(default: unit matrix)\n");
    printf("-Z       adaptable size                         "
                    "(default: fixed size)\n");
    printf("-N       normalize to unit integral             "
                    "(default: no normalization)\n");
    printf("-O       overwrite with option values           "
                    "(default: parameters read)\n");
    printf("-i#      initialization mode                    "
                    "(default: %s:%g:%g)\n", ininame, size, trim);
    printf("-w#      random offset range                    "
                    "(default: %g)\n", range);
    printf("-I#      initial cluster size (radius)          "
                    "(default: %g)\n", radius);
    printf("-S#      seed for random numbers                "
                    "(default: time)\n");
    printf("-e#      maximum number of update epochs        "
                    "(default: %"DIMID_FMT")\n", epochs);
    printf("-k#      patterns between two updates           "
                    "(default: %"DIMID_FMT")\n", update);
    printf("-s       do not shuffle data points             "
                    "(default: once per epoch)\n");
    printf("-q       do not normalize input ranges          "
                    "(default: mean=0, var=1)\n");
    printf("-a#      parameter update method                "
                    "(default: %s)\n", updname);
    printf("-t#:#:#  learning rates                         "
                    "(default: %g:%g:%g)\n",
                     lrates[0], lrates[1], lrates[2]);
    printf("-m#      momentum coefficient                   "
                    "(default: %g)\n", moment);
    printf("-g#:#    growth and shrink factor               "
                    "(default: %g:%g)\n", growth, shrink);
    printf("-z#:#    minimal and maximal change/lrate       "
                    "(default: %g:%g)\n", minchg, maxchg);
    printf("-y#      weight decay factor                    "
                    "(default: %g)\n", decay);
    printf("-T#      error for termination                  "
                    "(default: %g)\n", term);
    printf("-P#      error change for termination           "
                    "(default: %g)\n", delta);
    printf("-E       do not recompute error at end of epoch\n");
    printf("-H#      shape regularization parameter         "
                    "(default: %g)\n", regps[3]);
    printf("-R#:#:#  size  regularization parameters        "
                    "(default: %g:%g:%g)\n",
                     regps[0], regps[1], regps[2]);
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-u#      verbose output (print sse every # epochs)\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("domfile  file containing domain descriptions\n"
           "         (and maybe a pretrained network)\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("rbffile  file to write radial basis function network to\n");
    printf("infile   file to read  radial basis function network from "
                    "(only with -M)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get a default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                              break;
          case 'M': matinp  = 1;                         break;
          case 'o': optarg  = &trgname;                  break;
          case 'c': hidcnt  = (DIMID)strtol(s, &s, 0);   break;
          case 'G': gauss   = 0;                         break;
          case 'p': getdblvec(s, &s, 2, rfnps);          break;
          case 'v': type   |= CLS_VARS;                  break;
          case 'V': type   |= CLS_COVARS;                break;
          case 'Z': type   |= CLS_SIZE;                  break;
          case 'N': type   |= CLS_NORM;                  break;
          case 'O': owrite  = 1;                         break;
          case 'i': optarg  = &ininame;                  break;
          case 'w': range   =        strtod(s, &s);      break;
          case 'I': radius  =        strtod(s, &s);      break;
          case 'S': seed    =        strtol(s, &s, 0);   break;
          case 'e': epochs  = (DIMID)strtol(s, &s, 0);   break;
          case 'k': update  = (DIMID)strtol(s, &s, 0);   break;
          case 's': shuffle = 0;                         break;
          case 'q': norm    = 0;                         break;
          case 'a': optarg  = &updname;                  break;
          case 't': getdblvec(s, &s, 3, lrates);         break;
          case 'm': moment  =      strtod(s, &s);        break;
          case 'g': getdbls(s, &s, 2, &growth, &shrink); break;
          case 'z': getdbls(s, &s, 2, &minchg, &maxchg); break;
          case 'y': decay   =      strtod(s, &s);        break;
          case 'T': term    =      strtod(s, &s);        break;
          case 'H': getdbls(s, &s, 1, regps+3);          break;
          case 'R': getdblvec(s, &s, 3, regps);          break;
          case 'l': maxlen  = (int)  strtol(s, &s, 0);   break;
          case 'u': verbose = (DIMID)strtol(s, &s, 0);   break;
          case 'P': delta   =        strtod(s, &s);      break;
          case 'E': recomp  = 0;                         break;
          case 'r': optarg  = &recseps;                  break;
          case 'f': optarg  = &fldseps;                  break;
          case 'b': optarg  = &blanks;                   break;
          case 'C': optarg  = &comment;                  break;
          case 'd': mode   |= AS_DFLT;                   break;
          case 'h': optarg  = &fn_hdr;                   break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_rbf = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (matinp) {                 /* if matrix version */
    if ((k != 2) && (k != 3))   /* check the number */
      error(E_ARGCNT);          /* of arguments */
    fn_inp = fn_rbf;            /* shift the file names */
    fn_rbf = fn_tab;            /* (as there is no domain file) */
    fn_tab = fn_dom; fn_dom = NULL;
    if ((!fn_tab || !*fn_tab) && (!fn_inp || !*fn_inp))
      error(E_STDIN); }         /* stdin must not be used twice */
  else {                        /* if table version */
    if (k != 3) error(E_ARGCNT);/* check number of arguments */
    if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
      fn_hdr = "";              /* convert "-" to "" */
    i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
    if  (!fn_dom || !*fn_dom) i++;
    if  (!fn_tab || !*fn_tab) i++;
    if (i > 1) error(E_STDIN);  /* check assignments of stdin: */
  }                             /* stdin must not be used twice */
  s = strchr(ininame, ':');     /* check for a init. parameter */
  if (s != NULL) {              /* if such a parameter is given, */
    *s++ = 0;                   /* split into name and parameter */
    size = strtod(s, &end);     /* and get the parameter values */
    if (*end == ':') trim = strtod(s = end+1, &end);
    if (*end) error(E_INIPAR, s);
  }                             /* check for trailing info. */
  for (i = 0; initab[i].name; i++)
    if (strcmp(initab[i].name, ininame) == 0) break;
  if (!initab[i].name) error(E_METHOD, ininame);
  inimode = initab[i].code;     /* code the initialization mode */
  for (i = 0; updtab[i].name; i++)
    if (strcmp(updtab[i].name, updname) == 0) break;
  if (!updtab[i].name) error(E_METHOD, updname);
  method = updtab[i].code;      /* code the update method */
  if      (method == RBF_ADAPTIVE)  clsupd = CLS_ADAPTIVE;
  else if (method == RBF_RESILIENT) clsupd = CLS_RESILIENT;
  else if (method == RBF_QUICK)     clsupd = CLS_QUICK;
  else if (moment >  0)             clsupd = CLS_MOMENTUM;
  else                              clsupd = CLS_NONE;
  clsupd |= CLS_RBFGRAD;        /* code the custer update method */
  if (outcnt    <  0) error(E_UNITS,  OUTPUT);
  if (hidcnt    <  0) error(E_UNITS,  HIDDEN);
  if (lrates[0] <  0) error(E_LRATE,  lrates[0]);
  if (lrates[1] <  0) error(E_LRATE,  lrates[1]);
  if (lrates[2] <  0) error(E_LRATE,  lrates[2]);
  if (moment    <  0) error(E_MOMENT, moment);
  if (minchg    <  0) error(E_LPARAM, minchg);
  if (maxchg    <= 0) error(E_LPARAM, maxchg);
  if (growth    <  1) error(E_LPARAM, growth);
  if (shrink    >  1) error(E_LPARAM, shrink);
  if ((decay    <  0) || (decay >= 1)) error(E_LPARAM, decay);
  if (epochs    <  0) error(E_EPOCHS, epochs);
  rseed((unsigned)seed);        /* init. the random number generator */
  fputc('\n', stderr);          /* terminate the startup message */

  if (matinp) {                 /* if matrix version */
    /* --- parse radial basis function network --- */
    if (!fn_inp) m = -1;        /* if no input, clear att. counter */
    else {                      /* if an input file is given */
      t    = clock();           /* start the timer */
      scan = scn_create();      /* create a scanner */
      if (!scan) error(E_NOMEM);/* for the input network */
      if (scn_open(scan, NULL, fn_inp) != 0)
        error(E_FOPEN, scn_name(scan));
      fprintf(stderr, "reading %s ... ", scn_name(scan));
      clset = cls_parse(scan, 0, 1);    /* parse the cluster set */
      if (!clset) error(E_PARSE, scn_name(scan));
      if (!scn_eof(scan, 0)) {  /* if there is an input network, */
        rbf = rbf_parse(scan, clset, 0, 1);          /* parse it */
        if (!rbf || !scn_eof(scan, 1)) error(E_PARSE, scn_name(scan));
      }                         /* check for end of file */
      scn_delete(scan, 1);      /* delete the scanner and */
      scan   = NULL;            /* clear the scanner variable */
      inpcnt = cls_inpcnt(clset); /* get number of inputs/outputs */
      if (rbf) outcnt = rbf_outcnt(rbf);
      m = inpcnt +outcnt;       /* compute the number of values */
      if (m < 1) error(E_INPCNT); /* check for at least one input */
      if (rbf) {                /* if an RBF network is given */
        fprintf(stderr, "[%"DIMID_FMT" unit(s),",    rbf_unitcnt(rbf));
        fprintf(stderr, " %"DIMID_FMT" weight(s)]",  rbf_wgtcnt(rbf)); }
      else {                    /* if only a cluster set is given */
        fprintf(stderr, "[%"DIMID_FMT" input(s),",   inpcnt);
        fprintf(stderr, " %"DIMID_FMT" cluster(s)]", cls_clscnt(clset));
      }                         /* print a success message */
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    } }                         /* (depending on input) */

  else {                        /* if table version */
    /* --- parse domain descriptions --- */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);/* create an attribute set */
    scan = scn_create();        /* create a scanner */
    if (!scan)   error(E_NOMEM);/* for the domain file */
    t = clock();                /* start timer, open input file */
    if (scn_open(scan, NULL, fn_dom) != 0)
      error(E_FOPEN, scn_name(scan));
    fprintf(stderr, "reading %s ... ", scn_name(scan));
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan));
    a = as_attcnt(attset);      /* parse domain descriptions and */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */

    /* --- get target attribute --- */
    if (!scn_eof(scan, 0)) trgname = NULL;
    c = as_target(attset, trgname, +1);
    if (c < 0) error(trgname ? E_UNKTRG : E_MULTRG, trgname);
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    am_target(attmap, c);       /* and set the target attribute */
    inpcnt = am_inpcnt(attmap); /* get the number of inputs */
    outcnt = am_outcnt(attmap); /* and the number of outputs and */
    m      = inpcnt +outcnt;    /* the total number of variables */
    if (m < 1) error(E_INPCNT); /* check for at least one input */

    /* --- parse cluster set/RBF network --- */
    if (scn_eof(scan, 0))       /* if there is no input network */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
    else if ((scn_token(scan) == T_ID)/* if there is an RBF network */
    && ((strcmp(scn_value(scan), "rbfnet") == 0)
    ||  (strcmp(scn_value(scan), "rbf")    == 0))) {
      mode |= AS_NOXVAL;        /* prevent nominal domain extensions */
      rbf   = rbf_parse(scan, NULL, 0, 1);   /* parse neural network */
      if (!rbf || !scn_eof(scan, 1)) error(E_PARSE, scn_name(scan));
      clset = rbf_clset(rbf);   /* retrieve the cluster set */
      fprintf(stderr, "[%"DIMID_FMT" unit(s),",   rbf_unitcnt(rbf));
      fprintf(stderr, " %"DIMID_FMT" weight(s)]", rbf_wgtcnt (rbf)); }
    else {                      /* if there is a cluster set */
      mode |= AS_NOXVAL;        /* prevent nominal domain extensions */
      clset = cls_parse(scan, 0, 1);         /* parse cluster set */
      if (!clset || !scn_eof(scan, 1)) error(E_PARSE, scn_name(scan));
      fprintf(stderr, "[%"ATTID_FMT" input(s),",   cls_inpcnt(clset));
      fprintf(stderr, " %"DIMID_FMT" cluster(s)]", cls_clscnt(clset));
    }                           /* print a success message */
    scn_delete(scan, 1);        /* delete the scanner and */
    scan = NULL;                /* clear the scanner variable */
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }

  if (matinp) {                 /* if matrix version */
    /* --- read training patterns --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = mat_readx(&data, tread, 0, m);
    if (k) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* read the training patterns, */
    tread = NULL;               /* then close the input file */
    m = mat_colcnt(data);       /* get the number of variables */
    n = mat_rowcnt(data);       /* and the number of patterns */
    w = (double)n;              /* (which is also the total weight) */
    fprintf(stderr, "[%"DIMID_FMT" variable(s),", m);
    fprintf(stderr, " %"DIMID_FMT" pattern(s)]",  n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    if (m < 1) error(E_INPCNT); /* check for at least one input */
    if (n < 1) error(E_TPLCNT); /* and   for at least one data point */

    /* --- get number of inputs/outputs --- */
    if (trgname) {              /* if number of outputs given */
      outcnt = (DIMID)strtol(trgname, &s, 0);
      if ((s == trgname) || *s || (outcnt <= 0) || (outcnt >= m))
        error(E_UNITS, OUTPUT); /* get number of output variables */
    }                           /* (corresp. to target attribute) */
    inpcnt = m -outcnt;         /* compute the number of inputs */
  }  /* if (matinp) .. */

  else {                        /* if table version */
    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", a-1);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- read data table --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* create a data table */
    k = tab_read(table, tread, mode);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* read the table body and */
    tread = NULL;               /* delete the table reader */
    a = tab_colcnt(table);      /* get the number of attributes */
    n = tab_tplcnt(table);      /* and the number of data tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s),", a-1);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    if (n < 1) error(E_TPLCNT); /* check for at least one tuple */

    /* --- convert into matrix --- */
    data = tab_matrix(table, attmap, AM_BOTH, 1);
    if (!data) error(E_NOMEM);  /* convert the read table */
    table = NULL;               /* into a (data) matrix */
  }  /* if (matinp) .. else .. */

  /* --- create cluster set --- */
  if (!clset) {                 /* if no input cluster set is given */
    t = clock();                /* start the timer, print message */
    fprintf(stderr, "creating cluster set ... ");
    if (inpcnt <= 0) error(E_UNITS, INPUT);
    clset = cls_create(inpcnt, hidcnt);
    if (!clset) error(E_NOMEM); /* create a cluster set */
    cls_type(clset, type, radius *radius);
    for (r = 0; r < n; r++)     /* determine the ranges of values */
      cls_reg(clset, mat_row(data, r), 1);
    if (norm)                   /* if to normalize the data, */
      cls_reg(clset, NULL, 0);  /* compute the scaling factors */
    if (inimode < CLS_POINTS)   /* initialize the clusters */
      cls_init(clset, inimode, drand, NULL);
    else {                      /* or select initial centers */
      ctrs = mat_create(hidcnt, m);
      if (!ctrs                 /* create a matrix of cluster centers */
      || (cls_ctrsel(ctrs, data, inimode,size,trim, NULL,0, drand) < 0))
        error(E_NOMEM);         /* traverse the matrix rows */
      for (r = 0; r < hidcnt; r++)
        cls_init(clset, CLS_POINTS, drand, mat_row(ctrs, r));
      mat_delete(ctrs); ctrs = NULL;
    }                           /* set each matrix row as a center */
    if (range > 0)              /* jitter initial cluster centers */
      cls_jitter(clset, range, drand);
    owrite = -1;                /* set cluster type and radius */
    fprintf(stderr, "[%"DIMID_FMT" cluster(s)]", hidcnt);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a success message */
  if (owrite) {                 /* if to overwrite the parameters, */
    cls_type(clset, type, -1);  /* set the cluster parameters */
    if (gauss) cls_radfn(clset, rf_gauss,  rfd_gauss,  rfnps);
    else       cls_radfn(clset, rf_cauchy, rfd_cauchy, rfnps);
  }                             /* set the radial function */
  cls_norm(clset, CLS_NONE, NULL);
  cls_fwexp(clset, 0);          /* remove special features */
  cls_setup(clset);             /* and set up the cluster set */

  /* --- create radial basis function network --- */
  if (!rbf) {                   /* if no input network is given */
    t = clock();                /* start the timer, print message */
    fprintf(stderr, "creating network ... ");
    rbf = rbf_create(clset, outcnt);
    if (!rbf) error(E_NOMEM);   /* create an RBF network */
    for (r = 0; r < n; r++) {   /* traverse the training patterns */
      vec = mat_row(data, r);
      rbf_reg(rbf, vec, vec +inpcnt, 1);
    }                           /* register the input/target pairs */
    rbf_init(rbf);              /* initialize the connection weights */
    fprintf(stderr, "[%"DIMID_FMT" unit(s),",   rbf_unitcnt(rbf));
    fprintf(stderr, " %"DIMID_FMT" weight(s)]", rbf_wgtcnt(rbf));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a success message */
  /* --- set up radial basis function network for training --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "training network ... ");
  clset = rbf_clset(rbf);       /* get the cluster set */
  cls_method (clset, clsupd);   /* set the parameter update method */
  cls_lrate  (clset, lrates, NULL);   /* set the learning rates, */
  cls_moment (clset, moment);         /* the momentum coefficient, */
  cls_factors(clset, growth, shrink); /* the update factors and */
  cls_limits (clset, minchg, maxchg); /* the update limits */
  cls_regular(clset, regps);    /* set the regularization parameters */
  rbf_method (rbf, method);     /* set the parameter update method */
  rbf_lrate  (rbf, lrates[2]);  /* set the learning rate and */
  rbf_moment (rbf, moment);     /* the momentum coefficient */
  rbf_factors(rbf, growth, shrink);   /* set the update factors */
  rbf_limits (rbf, minchg, maxchg);   /* and the update limits */
  rbf_decay  (rbf, decay);      /* set the weight decay factor */
  rbf_setup  (rbf);             /* set up network for training */

  /* --- compute sse of untrained network --- */
  sse = INFINITY;               /* initialize the sum of errors */
  if (verbose) {                /* if verbose output is requested */
    sse = getsse(rbf, data);    /* compute the sum of errors */
    fprintf(stderr, "%15g\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", sse);
  }                             /* report initial sum of errors */
  sseold = sse;                 /* initialize old sum of errors */

  /* --- train radial basis function network --- */
  u = update; v = 0;            /* and initialize the counters */
  for (e = 0; e < epochs; e++){ /* do "epochs" epochs of training */
    if (shuffle)                /* shuffle the training patterns */
      mat_shuffle(data, drand, 0, 1);          /* (if requested) */
    for (sse = 0, r = mat_rowcnt(data); --r >= 0; ) {
      vec = mat_row(data, r);   /* traverse the training patterns */
      rbf_exec(rbf, vec, NULL); /* execute the neural network and */
      sse += rbf_bkprop(rbf, vec +inpcnt);  /* do error backprop. */
      if ((update > 0) && (--u <= 0)) {
        u = update; rbf_update(rbf); }
    }                           /* update after 'update' patterns */
    if (update <= 0)            /* if no number of patterns is given, */
      rbf_update(rbf);          /* update once in each epoch */
    if ((term >= 0)             /* if termination error set or */
    || (verbose && (--v <= 0))){/* if a verbose output is requested, */
      if (recomp)               /* compute error of current network */
        sse = getsse(rbf, data);
      if (verbose)              /* if verbose output requested */
        fprintf(stderr, "%15g\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", sse);
      v = verbose;              /* print sum of (squared) errors */
      if ((sse <= term)         /* if error or change is small enough */
      ||  (fabs(sse -sseold) <= delta))
        break;                  /* abort the training */
      sseold = sse;             /* note the current error */
    }                           /* (for comparisons) */
  }
  if (verbose)                  /* clear verbose error output */
    fprintf(stderr, "               \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");

  /* --- compute sse of trained network --- */
  sse = getsse(rbf, data);      /* compute the sum of errors */
  fprintf(stderr, "[%"DIMID_FMT" epoch(s)]", e);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- describe radial basis function network --- */
  t = clock();                  /* start the timer */
  if (fn_rbf && (strcmp(fn_rbf, "-") == 0)) fn_rbf = "";
  if (fn_rbf && *fn_rbf) { out = fopen(fn_rbf, "w"); }
  else                   { out = stdout; fn_rbf = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_rbf);
  if (!out) error(E_FOPEN, fn_rbf);
  if (!matinp) {                /* if table version */
    if (as_desc(attset, out, AS_TITLE|AS_IVALS|AS_DIRS, maxlen) != 0)
      error(E_FWRITE, fn_rbf);  /* describe attribute domains */
    fprintf(out, "\n");         /* leave one line empty */
  }
  if (rbf_desc(rbf, out, RBF_TITLE|RBF_INFO, maxlen) != 0)
    error(E_FWRITE, fn_rbf);    /* describe the RBF network */
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_rbf);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[sse: %g] ", sse);
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
